;;;; -*- lisp -*-

(in-package :yaclml)

(defvar *yaclml-stream* t)

(defmacro with-yaclml-stream (stream &body body)
  `(let ((*yaclml-stream* ,stream))
     (declare (special *yaclml-stream*))
     ,@body))

(defvar %yaclml-code% nil)

(defvar *expanders* (make-hash-table :test 'eql))

(defvar *expander-macros* (make-hash-table :test 'eql))

(defun yaclml-constant-p (thing)
  "Returns T if THING is, as far as yaclml is concerned, a run time
  constant."
  (or (stringp thing)
      (characterp thing)
      (numberp thing)
      (keywordp thing)))

(defun emit-princ (&rest items)
  "Emit to the current yaclml-code a form which will, at runtime,
princ ITEM. If (yaclml-constant-p ITEM) is true the princ will be done
at compile time."
  (dolist (item items)
    (if (yaclml-constant-p item)
        (push (princ-to-string item) %yaclml-code%)
        (push `(princ ,item *yaclml-stream*) %yaclml-code%))))

(defun emit-html (&rest items)
  "Like EMIT-PRINC but escapes html chars in item."
  (dolist (item items)
    (if (yaclml-constant-p item)
        (push (escape-as-html (princ-to-string item)) %yaclml-code%)
        (push `(princ (escape-as-html (princ-to-string ,item)) *yaclml-stream*) %yaclml-code%))))
  
(defun emit-code (&rest forms)
  "Emit to the current yaclml-code CODE. This means that whatever CODE
is it will be run, and it's result will be ignored, at runtime."
  (setf %yaclml-code% (nconc forms %yaclml-code%)))

(defun emit-princ-attributes (attributes)
  "Assuming attributes is an alist of name value pairs emit the code
nesseccary to print them at runtime. The attributes will be reordered
to get the longest possible constant string sequence.

If the value of any of the attributes is (the constant) NIL it will be
ignored.

If the value of any of the attributes is (the constant) T it the
=\"T\" will not be ignored. This is not valid XHTML, but it happens in
practice."
  (let ((constants '())
        (runtime '()))
    ;; we move the attributes around in order to get the longest constant sequence we can
    (dolist (attr attributes)
      (when (cdr attr) ;; this implies that passed attributes with a nil value are not printed
        (if (yaclml-constant-p (cdr attr))
            (push attr constants)
            (push attr runtime))))
    (dolist (attr constants)
      (emit-princ " " (car attr) "=\"")
      (if (eql t (cdr attr))
          ;; according to xhtml thoses attributes which in html are
          ;; specified without a value should just use the attribute
          ;; name as the xhtml value
          (emit-princ (car attr))
          (emit-princ (cdr attr)))
      (emit-princ "\""))
    (dolist (attr runtime)
      (emit-princ " " (car attr) "=\"")
      (emit-princ (cdr attr))
      (emit-princ "\""))))

(defun emit-open-tag (name attributes)
  "Emit the code required to print an open tag whose name is NAME and
with the attributes attributes."
  (emit-princ "<" name)
  (emit-princ-attributes attributes)
  (emit-princ ">"))

(defun emit-close-tag (name)
  "Emit the code required to print a close tag whose name is NAME."
  (emit-princ "</" name ">"))

(defun emit-empty-tag (name attributes)
  "Emit the code required to print an empty tag with name NAME and a
attributes ATTRIBUTES."
  (emit-princ "<" name)
  (emit-princ-attributes attributes)
  (emit-princ "/>"))

(defun emit-body (body)
  "Traverse body and emit the corresponding code. Every form in body
is analyzed according to the following rules:

cons whose car is not a known expander - code which should be included
with no further analysis.

cons whose car is a known expander - simply call the expander function
with the cdr of the cons as the arg.

yaclml-constant-p - print the constant (after escape-as-html) to
*yaclml-stream*.

cons whose car is YACLML-QUOTE - emit-body on every element of the
cdr.
"
  (dolist (form body)
    (emit-form form)))

(defun emit-form (form)
  (if (consp form)
      (let ((op (car form)))
        (cond
          ((gethash op *expander-macros*)
           (emit-form (funcall (gethash op *expander-macros*) (cdr form))))
          ((gethash op *expanders*)
           (funcall (gethash op *expanders*) (cdr form)))
          ((eql 'yaclml-quote op)
           (dolist (b (cdr form))
             (emit-form b)))
          ((eql 'cl:progn op)
           (dolist (b (cdr form))
             (emit-form b)))
          (t (emit-code form))))
      (if (yaclml-constant-p form)
          (emit-princ (escape-as-html (princ-to-string form)))
          (emit-code form))))

(defmacro deftag (name attributes &body body)
  "Define a new tag. ATTRIBUTES should be an attribute-spec (see
parse-attributes and attribute-bind. body is the body of the expander
lambda.

Within the BODY the functions EMIT-CODE, EMIT-PRINC and EMIT-HTML can
be used to generate code. EMIT-CODE should be passed lisp code which
will be executed at runtime.

Defines a macro named NAME which will setup the special variable
yaclml-code."
  (let ((contents (gensym)))
    `(progn
       (setf (gethash ',name *expanders*)
             (lambda (,contents)
               (attribute-bind ,attributes ,contents
                 ,@body)))
       (defmacro ,name (&rest contents)
         (let ((%yaclml-code% nil))
           (declare (special %yaclml-code%))
            ;; build tag's body
	   (funcall (gethash ',name *expanders*) contents)
           (setf %yaclml-code% (nreverse %yaclml-code%))
	   ;; now that we've generated the code we can fold the
	   ;; strings in yaclml-code and princ them, leaving any other
	   ;; forms as they are.
           `(progn ,@(mapcar (lambda (form)
                               (if (stringp form)
                                   `(write-string ,form *yaclml-stream*)
				   form))
                             (fold-strings %yaclml-code%))))))))

(defmacro deftag-macro (name attributes &body body)
  (let ((contents (gensym)))
    `(progn
       (defmacro ,name (&rest ,contents)
         (attribute-bind ,attributes ,contents ,@body))
       (setf (gethash ',name *expander-macros*) (lambda (,contents)
                                                  (attribute-bind ,attributes ,contents ,@body)))
       ',name)))

(defmacro def-simple-xtag (name)
  "Convience macro for defing tags which accept any kind of attribute
and just wrap the body in an xml tag."
  `(deftag ,name (&allow-other-attributes other-attributes &body body)
     (if body
         (progn
           (emit-open-tag ,(string-downcase (string name)) other-attributes)
           (emit-body body)
           (emit-close-tag ,(string-downcase (string name))))
         (emit-empty-tag ,(string-downcase (string name)) other-attributes))))

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;;; Copyright (c) 2002-2003, Edward Marco Baringer
;;;; All rights reserved. 
;;;; 
;;;; Redistribution and use in source and binary forms, with or without
;;;; modification, are permitted provided that the following conditions are
;;;; met:
;;;; 
;;;;  - Redistributions of source code must retain the above copyright
;;;;    notice, this list of conditions and the following disclaimer.
;;;; 
;;;;  - Redistributions in binary form must reproduce the above copyright
;;;;    notice, this list of conditions and the following disclaimer in the
;;;;    documentation and/or other materials provided with the distribution.
;;;;
;;;;  - Neither the name of Edward Marco Baringer, nor BESE, nor the names
;;;;    of its contributors may be used to endorse or promote products
;;;;    derived from this software without specific prior written permission.
;;;; 
;;;; THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
;;;; "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
;;;; LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
;;;; A PARTICULAR PURPOSE ARE DISCLAIMED.  IN NO EVENT SHALL THE COPYRIGHT
;;;; OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
;;;; SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
;;;; LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
;;;; DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
;;;; THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
;;;; (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
;;;; OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
